
#include "c8051.h"

#define SYSCLK 24500000


typedef unsigned char bool;
enum { false, true };

unsigned char txbusy;


struct Queue {
	unsigned char pos;
	unsigned char len;
	unsigned char buf[8];
};


static struct Queue uart0_tx_queue;
static struct Queue uart0_rx_queue;


void qinit(struct Queue *q)
{
	q->pos = q->len = 0;
}
bool qget(struct Queue *q, unsigned char *c)
{
	if (q->len > 0) {
		*c = q->buf[q->pos];
		q->pos = (q->pos + 1) % sizeof(q->buf);
		q->len--;
		return 1;
	}
	return 0;
}
void qput(struct Queue *q, unsigned char c)
{
	if (q->len < sizeof(q->buf)) {
		q->buf[(q->pos + q->len) % sizeof(q->buf)] = c;
		q->len++;
	}
}


void intr_uart0() interrupt 4
{
	unsigned char c;
	while (1) {
		if (TI0) {
			TI0 = 0;
			txbusy = 0;
			if (qget(&uart0_tx_queue, &c)) {
				SBUF0 = c;
				txbusy = 1;
			}
		} else if (RI0) {
			RI0 = 0;
			c = SBUF0;
			qput(&uart0_rx_queue, c);
		} else {
			break;
		}
	}
}

bool uart0_get(unsigned char *c)
{
	bool re;
	ES0 = 0;
	re = qget(&uart0_rx_queue, c);
	ES0 = 1;
	return re;
}
void uart0_put(unsigned char c)
{
	ES0 = 0;
	qput(&uart0_tx_queue, c);
	if (!txbusy) {
		if (qget(&uart0_tx_queue, &c)) {
			SBUF0 = c;
			txbusy = 1;
			TI0 = 0;
		}
	}
	ES0 = 1;
}

void main()
{
	PCA0MD = 0x00;

	// initialize system clock
	OSCICN = 0x07;		// oscillator divided by 1
	RSTSRC = 0x04;

	// initialize I/O port
	XBR0 = 0x04;
	XBR1 = 0x03;
	XBR2 = 0x40;
	P0MDOUT |= 0x04;

	// initialize timer
	CKCON = 0x00;
	TMOD = 0x20;

	// timer1
	TH1 = -(SYSCLK / 2 / 12 / 9600);		// 9600bps
	TL1 = 0x00;
	TR1 = 1;

	// initialize uart0
	SCON0 = 0x50;
	RI0 = 0;
	TI0 = 0;
	ES0 = 1;									// enable interrupt
	txbusy = 0;

	//

	qinit(&uart0_tx_queue);
	qinit(&uart0_rx_queue);

	//

	EA = 1;

	while (1) {
		unsigned char c;

		if (uart0_get(&c)) {
			uart0_put(c);
		}
	}
}

